@@ -1,6 +1,6 @@
 #!/usr/bin/env python3
 # localslackirc
-# Copyright (C) 2018-2020 Salvo "LtWorf" Tomaselli
+# Copyright (C) 2018-2021 Salvo "LtWorf" Tomaselli
 #
 # localslackirc is free software: you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -34,6 +34,14 @@ from socket import gethostname
 import sys
 import time
 
+
+try:
+    from emoji import emojize
+except ModuleNotFoundError:
+    def emojize(string:str, use_aliases:bool=False, delimiters: Tuple[str,str]=(':', ':')) -> str:  # type: ignore
+        return string
+
+
 import slack
 from log import *
 
@@ -88,39 +96,58 @@ class Provider(Enum):
 MPIM_HIDE_DELAY = datetime.timedelta(days=50)
 
 
+class ClientSettings(NamedTuple):
+    nouserlist: bool
+    autojoin: bool
+    provider: Provider
+    ignored_channels: Set[bytes]
+    silenced_yellers: Set[bytes]
+    downloads_directory: Path
+    formatted_max_lines: int = 0
+
+    def verify(self) -> Optional[str]:
+        '''
+        Make sure that the configuration is correct.
+
+        In that case return None. Otherwise an error string.
+        '''
+        if not self.downloads_directory.is_dir():
+            return f'{self.downloads_directory} is not a directory'
+        return None
+
+
 class Client:
     def __init__(
                     self,
                     s: asyncio.streams.StreamWriter,
                     sl_client: Union[slack.Slack],
-                    nouserlist: bool,
-                    autojoin: bool,
-                    provider: Provider,
-                    ignored_channels: Set[bytes],
+                    settings: ClientSettings,
+
     ):
         self.nick = b''
         self.username = b''
         self.realname = b''
-        self.parted_channels: Set[bytes] = ignored_channels
+        self.parted_channels: Set[bytes] = settings.ignored_channels
         self.hostname = gethostname().encode('utf8')
 
-        self.provider = provider
+        self.settings = settings
         self.s = s
         self.sl_client = sl_client
-        self.nouserlist = nouserlist
-        self.autojoin = autojoin
         self._usersent = False # Used to hold all events until the IRC client sends the initial USER message
         self._held_events: List[slack.SlackEvent] = []
         self._mentions_regex_cache: Dict[str, Optional[re.Pattern]] = {}  # Cache for the regexp to perform mentions. Key is channel id
 
-        if self.provider == Provider.SLACK:
+        if self.settings.provider == Provider.SLACK:
             self.substitutions = _SLACK_SUBSTITUTIONS
         else:
             self.substitutions = []
 
     async def _nickhandler(self, cmd: bytes) -> None:
-        _, nick = cmd.split(b' ', 1)
-        self.nick = nick.strip()
+        if b' ' not in cmd:
+            self.nick = b'localslackirc'
+        else:
+            _, nick = cmd.split(b' ', 1)
+            self.nick = nick.strip()
         assert self.sl_client.login_info
         if self.nick != self.sl_client.login_info.self.name.encode('ascii'):
             await self._sendreply(Replies.ERR_ERRONEUSNICKNAME, 'Incorrect nickname, use %s' % self.sl_client.login_info.self.name)
@@ -133,7 +160,7 @@ class Client:
 
         extratokens.insert(0, self.nick)
 
-        self.s.write(b':%s %03d %s :%s\n' % (
+        self.s.write(b':%s %03d %s :%s\r\n' % (
             self.hostname,
             codeint,
             b' '.join(i if isinstance(i, bytes) else i.encode('utf8') for i in extratokens),
@@ -151,13 +178,12 @@ class Client:
         await self._sendreply(2, 'Your nickname must be: %s' % self.sl_client.login_info.self.name)
         await self._sendreply(Replies.RPL_LUSERCLIENT, 'There are 1 users and 0 services on 1 server')
 
-        if self.autojoin and not self.nouserlist:
+        if self.settings.autojoin and not self.settings.nouserlist:
             # We're about to load many users for each chan; instead of requesting each
             # profile on its own, batch load the full directory.
             await self.sl_client.prefetch_users()
 
-        if self.autojoin:
-
+        if self.settings.autojoin:
             mpim_cutoff = datetime.datetime.utcnow() - MPIM_HIDE_DELAY
 
             for sl_chan in await self.sl_client.channels():
@@ -186,7 +212,7 @@ class Client:
 
     async def _pinghandler(self, cmd: bytes) -> None:
         _, lbl = cmd.split(b' ', 1)
-        self.s.write(b':%s PONG %s %s\n' % (self.hostname, self.hostname, lbl))
+        self.s.write(b':%s PONG %s %s\r\n' % (self.hostname, self.hostname, lbl))
         await self.s.drain()
 
     async def _joinhandler(self, cmd: bytes) -> None:
@@ -214,7 +240,7 @@ class Client:
             await self._sendreply(Replies.ERR_NOSUCHCHANNEL, f'Unable to join channel: {channel_name}')
 
     async def _send_chan_info(self, channel_name: bytes, slchan: slack.Channel):
-        if not self.nouserlist:
+        if not self.settings.nouserlist:
             userlist: List[bytes] = []
             for i in await self.sl_client.get_members(slchan.id):
                 try:
@@ -230,15 +256,16 @@ class Client:
 
             users = b' '.join(userlist)
 
-        self.s.write(b':%s!salvo@127.0.0.1 JOIN %s\n' % (self.nick, channel_name))
+        self.s.write(b':%s!%s@127.0.0.1 JOIN %s\r\n' % (self.nick, self.nick, channel_name))
         await self.s.drain()
         await self._sendreply(Replies.RPL_TOPIC, slchan.real_topic, [channel_name])
-        await self._sendreply(Replies.RPL_NAMREPLY, b'' if self.nouserlist else users, ['=', channel_name])
+        await self._sendreply(Replies.RPL_NAMREPLY, b'' if self.settings.nouserlist else users, ['=', channel_name])
         await self._sendreply(Replies.RPL_ENDOFNAMES, 'End of NAMES list', [channel_name])
 
     async def _privmsghandler(self, cmd: bytes) -> None:
         _, dest, msg = cmd.split(b' ', 2)
-        msg = msg.lstrip(b':')
+        if msg.startswith(b':'):
+            msg = msg[1:]
 
         # Handle sending "/me does something"
         # b'PRIVMSG #much_private :\x01ACTION saluta tutti\x01'
@@ -340,8 +367,8 @@ class Client:
         users = cmd.split(b' ')
         del users[0]
 
-        if len(users) > 1:
-            await self._sendreply(Replies.ERR_UNKNOWNCOMMAND, 'Server parameter is not supported')
+        if len(users) != 1:
+            await self._sendreply(Replies.ERR_UNKNOWNCOMMAND, 'Syntax: /whois nickname')
             return
 
         # Seems that oftc only responds to the last one
@@ -452,7 +479,8 @@ class Client:
         await self._sendreply(Replies.RPL_ENDOFWHO, 'End of WHO list', [name])
 
     async def sendmsg(self, from_: bytes, to: bytes, message: bytes) -> None:
-        self.s.write(b':%s!salvo@127.0.0.1 PRIVMSG %s :%s\n' % (
+        self.s.write(b':%s!%s@127.0.0.1 PRIVMSG %s :%s\r\n' % (
+            from_,
             from_,
             to, #private message, or a channel
             message,
@@ -493,7 +521,7 @@ class Client:
         """
         for i in self.substitutions:
             msg = msg.replace(i[1], i[0])
-        if self.provider == Provider.SLACK:
+        if self.settings.provider == Provider.SLACK:
             msg = msg.replace('@here', '<!here>')
             msg = msg.replace('@channel', '<!channel>')
             msg = msg.replace('@everyone', '<!everyone>')
@@ -508,16 +536,13 @@ class Client:
             username = m.string[m.start():m.end()]
             if username.startswith('://'):
                 continue # Match inside a url
-            elif self.provider == Provider.SLACK:
+            elif self.settings.provider == Provider.SLACK:
                 msg = msg[0:m.start()] + '<@%s>' % (await self.sl_client.get_user_by_name(username)).id + msg[m.end():]
         return msg
 
-    async def parse_message(self, i: str) -> bytes:
+    async def parse_message(self, i: str, source: bytes) -> bytes:
         # Replace all mentions with @user
-        while True:
-            mention = _MENTIONS_REGEXP.search(i)
-            if not mention:
-                break
+        while mention := _MENTIONS_REGEXP.search(i):
             i = (
                 i[0:mention.span()[0]] +
                 (await self.sl_client.get_user(mention.groups()[0])).name +
@@ -525,39 +550,78 @@ class Client:
             )
 
         # Replace all channel mentions
-        if self.provider == Provider.SLACK:
-            while True:
-                mention = _CHANNEL_MENTIONS_REGEXP.search(i)
-                if not mention:
-                    break
-                i = (
-                    i[0:mention.span()[0]] +
-                    '#' +
-                    (await self.sl_client.get_channel(mention.groups()[0])).name_normalized +
-                    i[mention.span()[1]:]
-                )
+        while mention := _CHANNEL_MENTIONS_REGEXP.search(i):
+            i = (
+                i[0:mention.span()[0]] +
+                '#' +
+                (await self.sl_client.get_channel(mention.groups()[0])).name_normalized +
+                i[mention.span()[1]:]
+            )
 
-            while True:
-                url = _URL_REGEXP.search(i)
-                if not url:
-                    break
-                schema, path, label = url.groups()
+        #Replace URL with bottom links
+        bottom = ""
+        refs = str.maketrans("0123456789", "⁰¹²³⁴⁵⁶⁷⁸⁹")
+        refn = 1
+        while url := _URL_REGEXP.search(i):
+            schema, path, label = url.groups()
+            if label and (label != f"{schema}://{path}"):
+                ref = str(refn).translate(refs)
+                refn += 1
+                i = (
+                    i[0:url.span()[0]] +
+                    f'{label}{ref}' +
+                    i[url.span()[1]:]
+                )
+                bottom += f'\n  {ref} {schema}://{path}'
+            else:
                 i = (
                     i[0:url.span()[0]] +
                     f'{schema}://{path}' +
-                    (f' ({label})' if label else '') +
                     i[url.span()[1]:]
                 )
+        i += bottom
 
         for s in self.substitutions:
             i = i.replace(s[0], s[1])
 
+        # Replace emoji codes (e.g. :thumbsup:)
+        i = emojize(i, use_aliases=True)
+
+        # Store long formatted text into txt files
+        if self.settings.formatted_max_lines:
+            begin = 0
+            while True:
+                try:
+                    fmtstart = i.index('```', begin)
+                    fmtend = i.index('```', fmtstart + 1)
+
+                    formatted = i[fmtstart + 3:fmtend]
+                    prefix = i[0:fmtstart]
+                    suffix = i[fmtend + 3:]
+
+                    if formatted.count('\n') > self.settings.formatted_max_lines:
+                        import tempfile
+                        with tempfile.NamedTemporaryFile(
+                                    mode='wt',
+                                    dir=self.settings.downloads_directory,
+                                    suffix='.txt',
+                                    prefix='localslackirc-attachment-',
+                                    delete=False) as tmpfile:
+                            tmpfile.write(formatted)
+                            i = prefix + f'\n === PREFORMATTED TEXT AT file://{tmpfile.name}\n' + suffix
+                    else:
+                        i = '```'.join((prefix, formatted, suffix))
+                        begin = fmtend + 1
+                except ValueError:
+                    break
+
         encoded = i.encode('utf8')
 
-        if self.provider == Provider.SLACK:
-            encoded = encoded.replace(b'<!here>', b'yelling [%s]' % self.nick)
-            encoded = encoded.replace(b'<!channel>', b'YELLING LOUDER [%s]' % self.nick)
-            encoded = encoded.replace(b'<!everyone>', b'DEAFENING YELL [%s]' % self.nick)
+        if self.settings.provider == Provider.SLACK:
+            yell = b' [%s]:' % self.nick if source not in self.settings.silenced_yellers else b':'
+            encoded = encoded.replace(b'<!here>', b'yelling%s' % yell)
+            encoded = encoded.replace(b'<!channel>', b'YELLING LOUDER%s' % yell)
+            encoded = encoded.replace(b'<!everyone>', b'DEAFENING YELL%s' % yell)
 
         return encoded
 
@@ -580,10 +644,12 @@ class Client:
             # Ignoring messages, channel was left on IRC
             return
 
-        for msg in (prefix + sl_ev.text).split('\n'):
-            if not msg:
+        text = sl_ev.text
+
+        lines = await self.parse_message(prefix + text, source)
+        for i in lines.split(b'\n'):
+            if not i:
                 continue
-            i = await self.parse_message(msg)
             if isinstance(sl_ev, slack.ActionMessage):
                 i = b'\x01ACTION ' + i + b'\x01'
             await self.sendmsg(
@@ -612,17 +678,18 @@ class Client:
         name = user.name.encode('utf8')
         rname = user.real_name.replace(' ', '_').encode('utf8')
         if joined:
-            self.s.write(b':%s!%s@127.0.0.1 JOIN :%s\n' % (name, rname, dest))
+            self.s.write(b':%s!%s@127.0.0.1 JOIN :%s\r\n' % (name, rname, dest))
         else:
-            self.s.write(b':%s!%s@127.0.0.1 PART %s\n' % (name, rname, dest))
+            self.s.write(b':%s!%s@127.0.0.1 PART %s\r\n' % (name, rname, dest))
         await self.s.drain()
 
     async def slack_event(self, sl_ev: slack.SlackEvent) -> None:
         if not self._usersent:
             self._held_events.append(sl_ev)
             return
+
         if isinstance(sl_ev, slack.MessageDelete):
-            await self._message(sl_ev, '[deleted]')
+            await self._message(sl_ev, '[deleted] ')
         elif isinstance(sl_ev, slack.Message):
             await self._message(sl_ev)
         elif isinstance(sl_ev, slack.ActionMessage):
@@ -631,7 +698,7 @@ class Client:
             if sl_ev.is_changed:
                 await self._message(sl_ev.diffmsg)
         elif isinstance(sl_ev, slack.MessageBot):
-            await self._message(sl_ev, '[%s]' % sl_ev.username)
+            await self._message(sl_ev, '[%s] ' % sl_ev.username)
         elif isinstance(sl_ev, slack.FileShared):
             f = await self.sl_client.get_file(sl_ev)
             await self._message(f.announce())
@@ -728,14 +795,24 @@ def main() -> None:
                                 help='allow non 127. addresses, this is potentially dangerous')
     parser.add_argument('-f', '--status-file', type=str, action='store', dest='status_file', required=False, default=None,
                                 help='Path to the file to keep the internal status.')
+    parser.add_argument('-d', '--debug', action='store_true', dest='debug', required=False, default=False,
+                                help='Enables debugging logs.')
     parser.add_argument('--log-suffix', type=str, action='store', dest='log_suffix', default='',
                                 help='Set a suffix for the syslog identifier')
     parser.add_argument('--ignored-channels', type=str, action='store', dest='ignored_channels', default='',
                                 help='Comma separated list of channels to not join when autojoin is enabled')
+    parser.add_argument('--downloads-directory', type=str, action='store', dest='downloads_directory', default='/tmp',
+                                help='Where to create files for automatic downloads')
+    parser.add_argument('--formatted-max-lines', type=int, action='store', dest='formatted_max_lines', default=0,
+                                help='Maximum amount of lines in a formatted text to send to the client rather than store in a file.\n'
+                                'Setting to 0 (the default) will send everything to the client')
+    parser.add_argument('--silenced-yellers', type=str, action='store', dest='silenced_yellers', default='',
+                                help='Comma separated list of nicknames that won\'t generate notifications when using @channel and @here')
 
     args = parser.parse_args()
 
     openlog(environ.get('LOG_SUFFIX', args.log_suffix))
+    set_debug(environ.get('DEBUG', args.debug))
 
     status_file_str: Optional[str] = environ.get('STATUS_FILE', args.status_file)
     status_file = None
@@ -767,6 +844,22 @@ def main() -> None:
     else:
         ignored_channels = set()
 
+    if 'DOWNLOADS_DIRECTORY' in environ:
+        downloads_directory = Path(environ['DOWNLOADS_DIRECTORY'])
+    else:
+        downloads_directory = Path(args.downloads_directory)
+
+    try:
+        formatted_max_lines = int(environ.get('FORMATTED_MAX_LINES', args.formatted_max_lines))
+    except:
+        exit('FORMATTED_MAX_LINES is not a valid int')
+
+    yellers_str = environ.get('SILENCED_YELLERS', args.silenced_yellers)
+    if yellers_str:
+        silenced_yellers = {i.strip().encode('utf8') for i in yellers_str.split(',')}
+    else:
+        silenced_yellers = set()
+
 
     if 'TOKEN' in environ:
         token = environ['TOKEN']
@@ -819,7 +912,19 @@ def main() -> None:
         sl_client = slack.Slack(token, cookie, previous_status)
         await sl_client.login()
 
-        ircclient = Client(writer, sl_client, nouserlist, autojoin, provider, ignored_channels)
+        clientsettings = ClientSettings(
+            nouserlist=nouserlist,
+            autojoin=autojoin,
+            provider=provider,
+            ignored_channels=ignored_channels,
+            downloads_directory=downloads_directory,
+            formatted_max_lines=formatted_max_lines,
+            silenced_yellers=silenced_yellers,
+        )
+        verify = clientsettings.verify()
+        if verify is not None:
+            exit(verify)
+        ircclient = Client(writer, sl_client, clientsettings)
 
         try:
             from_irc_task = asyncio.create_task(from_irc(reader, ircclient))
