@@ -1,5 +1,5 @@
 # localslackirc
-# Copyright (C) 2018-2020 Salvo "LtWorf" Tomaselli
+# Copyright (C) 2018-2021 Salvo "LtWorf" Tomaselli
 #
 # localslackirc is free software: you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -32,6 +32,7 @@ from log import *
 
 
 USELESS_EVENTS = {
+    'accounts_changed',
     'channel_marked',
     'group_marked',
     'mpim_marked',
@@ -216,10 +217,24 @@ class FileShared:
 class MessageBot:
     type: Literal['message']
     subtype: Literal['bot_message']
-    text: str
+    _text: str = field(metadata={'name': 'text'})
     username: str
     channel: str
     bot_id: Optional[str] = None
+    attachments: List[Dict[str, Any]] = field(default_factory=list)
+
+    @property
+    def text(self):
+        r = [self._text]
+        for i in self.attachments:
+            t = ""
+            if 'text' in i:
+                t = i['text']
+            elif 'fallback' in i:
+                t = i['fallback']
+            for line in t.split("\n"):
+                r.append("| " + line)
+        return '\n'.join(r)
 
 
 class User(NamedTuple):
@@ -270,6 +285,7 @@ class HistoryBotMessage:
     ts: float = 0
     files: List[File] = field(default_factory=list)
     thread_ts: Optional[str] = None
+    attachments: List[Dict[str, Any]] = field(default_factory=list)
 
 
 @dataclass
@@ -405,10 +421,16 @@ class Slack:
         dt = datetime.datetime.fromtimestamp(last_timestamp)
         log(f'Last known timestamp {dt}')
 
-        for channel in await self.channels():
-            if not channel.is_member:
-                continue
-            log(f'Downloading logs from channel {channel.name_normalized}')
+        chats: Sequence[Union[IM, Channel]] = []
+        chats += await self.channels() + await self.get_ims()  # type: ignore
+        for channel in chats:
+            if isinstance(channel, Channel):
+                if not channel.is_member:
+                    continue
+
+                log(f'Downloading logs from channel {channel.name_normalized}')
+            else:
+                log(f'Downloading logs from IM {channel.user}')
 
             cursor = None
             while True: # Loop to iterate the cursor
@@ -459,7 +481,8 @@ class Slack:
                     elif isinstance(msg, HistoryBotMessage):
                         self._internalevents.append(MessageBot(
                             type='message', subtype='bot_message',
-                            text=msg.text,
+                            _text=msg.text,
+                            attachments=msg.attachments,
                             username=msg.username,
                             channel=channel.id,
                             bot_id=msg.bot_id,
@@ -817,18 +840,21 @@ class Slack:
             if t in USELESS_EVENTS:
                 continue
 
+            debug(event)
+            loadable_events = Union[TopicChange, FileShared, MessageBot, MessageEdit, MessageDelete, GroupJoined, Join, Leave]
             try:
-                ev: Optional[Union[TopicChange, FileShared, MessageBot, MessageEdit, MessageDelete, GroupJoined, Join, Leave]] = load(
+                ev: Optional[loadable_events] = load(
                     event,
-                    Union[TopicChange, FileShared, MessageBot, MessageEdit, MessageDelete, GroupJoined, Join, Leave]  # type: ignore
+                    loadable_events  # type: ignore
                 )
             except Exception:
                 ev = None
 
-            if isinstance(ev, Join):
-                self._get_members_cache[ev.channel].add(ev.user)
-            elif isinstance(ev, Leave):
-                self._get_members_cache[ev.channel].remove(ev.user)
+            if isinstance(ev, (Join, Leave)) and ev.channel in self._get_members_cache:
+                if isinstance(ev, Join):
+                    self._get_members_cache[ev.channel].add(ev.user)
+                else:
+                    self._get_members_cache[ev.channel].remove(ev.user)
 
             if ev:
                 return ev
